@@ -417,6 +417,20 @@ let lookup_direct pdf key dict =
       end
   | _ -> None
 
+let lookup_immediate key dict =
+  match dict with
+  | Dictionary d | Stream {contents = (Dictionary d, _)} -> lookup_string_compare key d
+  | _ -> None
+
+(* Follow a nested chain of dictionary entries. *)
+let rec lookup_chain pdf obj = function
+  | [] -> Some obj
+  | [n] -> lookup_direct pdf n obj
+  | n::ns ->
+      match lookup_direct pdf n obj with
+      | Some obj' -> lookup_chain pdf obj' ns
+      | None -> None
+
 let indirect_number pdf key dict =
   match direct pdf dict with
   | Dictionary d | Stream {contents = (Dictionary d, _)} ->
@@ -429,6 +443,55 @@ let indirect_number pdf key dict =
       end
   | _ -> None
 
+(* Add an object, given an object number. *)
+let addobj_given_num doc (num, obj) =
+  doc.objects.maxobjnum <-
+    max doc.objects.maxobjnum num;
+  doc.objects.pdfobjects <-
+    pdfobjmap_add num (ref (Parsed obj), 0) doc.objects.pdfobjects
+
+(* Follow a chain from the root, finding a dictionary entry to replace (or add).
+   Keep the same direct / indirect structure as is already present. NB this
+   cannot create a dictionary or tree of dictionaries which are not already
+   there, because the information about direct / indirect is not present. *)
+
+(* FIXME Unexpectedly complex! To be tested thoroughly before in public API. *)
+
+(* Find the final indirect object in the chain, returning its number and the
+   remaining (fully-direct) chain *)
+let rec find_final_indirect remaining_chain pdf obj objnum = function
+  | [] -> (objnum, rev remaining_chain)
+  | k::ks ->
+      match indirect_number pdf k obj with
+      | Some i -> find_final_indirect [] pdf (lookup_obj pdf i) i ks
+      | None ->
+          match lookup_immediate k obj with
+          | Some obj -> find_final_indirect (k::remaining_chain) pdf obj objnum ks
+          | None -> assert false (* chain pre-checked by lookup_chain *)
+
+(* The object number is now chosen. We follow what remains of the chain and insert
+the new key-value pair. *)
+let rec replace_chain_all_direct finalobj chain (k, v) =
+  match finalobj with
+  | Dictionary _ as d | Stream ({contents = (d, _)}) ->
+      begin match chain with
+      | [] -> add_dict_entry d k v
+      | c::cs -> add_dict_entry d c (replace_chain_all_direct finalobj cs (k, v))
+      end
+  | _ -> raise (PDFError "replace_chain_all_direct: bad chain")
+
+let replace_chain pdf chain (k, v) =
+  match lookup_chain pdf pdf.trailerdict chain with
+  | None -> raise (PDFError "chain must already exist")
+  | Some _ ->
+      match chain with
+      | [] -> raise (PDFError "no chain")
+      | chain ->
+          let finalobjnum, remaining_chain = find_final_indirect [] pdf pdf.trailerdict 0 chain in
+            if finalobjnum = 0 then raise (PDFError "cannot use replace_chain to set trailer dictonary") else
+              let newobj = replace_chain_all_direct (lookup_obj pdf finalobjnum) remaining_chain (k, v) in
+                addobj_given_num pdf (finalobjnum, newobj)
+
 (* Look up under a key and its alternate. Return the value associated
 with the key that worked, or [None] if neither did. *)
 let lookup_direct_orelse pdf k k' d =
@@ -546,12 +609,6 @@ let list_of_objs doc =
     objiter (fun k v -> objs =| (k, Parsed v)) doc;
     !objs
 
-(* Add an object, given an object number. *)
-let addobj_given_num doc (num, obj) =
-  doc.objects.maxobjnum <-
-    max doc.objects.maxobjnum num;
-  doc.objects.pdfobjects <-
-    pdfobjmap_add num (ref (Parsed obj), 0) doc.objects.pdfobjects
 
 (* Add an object. We use the first number larger than the maxobjnum,
 and update that. *)
