@@ -127,9 +127,19 @@ val lookup_fail : string -> (t -> string -> pdfobject -> pdfobject)
 (** Same, but with customised exception. *)
 val lookup_exception : exn -> t -> string -> pdfobject -> pdfobject
 
-(** [lookup_direct doc key dict] looks up the key returning an option type. *) 
+(** [lookup_direct doc key dict] looks up the key, resolving indirections at
+    source and destination, returning an option type. *) 
 val lookup_direct : t -> string -> pdfobject -> pdfobject option
 
+(** [lookup_immediate key dict] looks up the key returning the value, without
+following indirects at either source or destination. *)
+val lookup_immediate : string -> pdfobject -> pdfobject option
+
+(** [lookup_chain doc start keys] looks up the key in a nested dictionary. For
+    example [lookup_chain pdf pdf.Pdf.trailerdict ["/Root"; "/StructTreeRoot";
+    "/RoleMap"]] *)
+val lookup_chain : t -> pdfobject -> string list -> pdfobject option
+
 (** Return the object number of an indirect dictionary object, if it is indirect. *)
 val indirect_number : t -> string -> pdfobject -> int option
 
@@ -285,6 +295,10 @@ val change_id : t -> string -> unit
 
 (**/**)
 
+val replace_chain : t -> string list -> string * pdfobject -> unit
+
+val replace_chain_all_direct : pdfobject -> string list -> string * pdfobject -> pdfobject
+
 (* This is only for the use of Pdfread for when the /Length is incorrect. *)
 type toget_crypt =
   | NoChange
