@@ -28,7 +28,7 @@ convert this to the output stream. *)
 let encode_ASCIIHex stream =
   let size = bytes_size stream in
     let stream' = mkbytes (size * 2 + 1) in
-      bset stream' (size * 2) (int_of_char '>'); (*r ['>'] is end-of-data *)
+      bset stream' (size * 2) (int_of_char '>'); (* ['>'] is end-of-data *)
       for p = 0 to size - 1 do
         let chars = explode (Printf.sprintf "%02X" (bget stream p)) in
           bset stream' (p * 2) (int_of_char (hd chars));
@@ -293,8 +293,8 @@ let decode_flate stream =
 
 (* Decode LZW. *)
 let decode_lzw early i =
-  let prefix_code = Array.make 4096 0
-  and append_character = Array.make 4096 0
+  let prefix_code = Array.make 8192 0
+  and append_character = Array.make 8192 0
   and bit_count = ref 0
   and bit_buffer = ref 0l
   and endflush = ref 4
@@ -967,14 +967,27 @@ let decode_predictor pred colors bpc columns stream =
           done;
           bytes_of_arraylist (rev !outputlines)
 
-(* Assumes 12 (PNGUp, no column padding required, colors, bpc standard values.
-This is just for xref streams for now *)
+(* A couple of special-purpose encoders for prediction. *)
 let encode_predictor pred colors bpc columns stream =
-  if pred = 12 then
-    begin
-      let get0 a i =
-        if i < 0 then 0 else bget a i
-      in
+  let get0 a i =
+    if i < 0 then 0 else bget a i
+  in
+    match pred with
+    | 11 ->
+        (* Just for recompressing inline images. Restricted to bpc = 8, colors = 3 for now. *)
+        let o, bytes = Pdfio.input_output_of_bytes 4096 in
+          for scanline = 0 to bytes_size stream / (columns * 3) - 1 do
+            o.Pdfio.output_byte 1; (* tag for Sub *)
+            for byte = 0 to columns * 3 - 1 do
+              o.Pdfio.output_byte
+                ((get0 stream (scanline * columns + byte)) -
+                 (get0 stream (scanline * columns + byte - 3))
+                 mod 256)
+            done
+          done;
+          Pdfio.extract_bytes_from_input_output o bytes
+    | 12 ->
+        (* Just for XRef streams. *)
         let o, bytes = Pdfio.input_output_of_bytes 4096 in
           for scanline = 0 to bytes_size stream / columns - 1 do
             o.Pdfio.output_byte 2; (* tag for Up *)
@@ -985,10 +998,9 @@ let encode_predictor pred colors bpc columns stream =
                  mod 256)
             done
           done;
-      Pdfio.extract_bytes_from_input_output o bytes
-    end
-  else
-    raise (Pdf.PDFError "encode_predictor: not supported")
+        Pdfio.extract_bytes_from_input_output o bytes
+    | n ->
+        raise (Pdf.PDFError ("encode_predictor: not supported - " ^ string_of_int n))
 
 (* Run Length Encoding *)
 let encode_runlength stream =
@@ -1409,7 +1421,7 @@ let process_prediction predictor predictor_columns stream =
   | _ -> assert false
 
 (* Encode a PDF stream with an encoding. *)
-let encode_pdfstream pdf encoding ?predictor ?(predictor_columns = 1) stream =
+let encode_pdfstream pdf encoding ?(only_if_smaller=false) ?predictor ?(predictor_columns = 1) stream =
   Pdf.getstream stream;
   match stream with
   | Pdf.Stream ({contents = d, Pdf.Got s} as stream) ->
@@ -1420,5 +1432,5 @@ let encode_pdfstream pdf encoding ?predictor ?(predictor_columns = 1) stream =
       in
         let data = encoder_of_encoding encoding predicted in
           let d'' = add_encoding (bytes_size data) pdf encoding d' in
-            stream := d'', Pdf.Got data
+            if not only_if_smaller || bytes_size data + 20 < bytes_size s then stream := d'', Pdf.Got data
   | _ -> raise (Pdf.PDFError "Pdf.encode_pdfstream: malformed Stream")
