@@ -84,104 +84,6 @@ let merge_pdfs_renumber names pdfs =
         let table = combine first_names (Pdf.renumber_pdfs first_pdfs) in
           map (function k -> lookup_failnull k table) names
       
-(* Reading a name tree, flattened. FIXME: This appears to be somewhat duplicated in Pdf.ml.
-   Move name / number tree stuff to pdftree.ml *)
-let rec read_name_tree pdf tree =
-  let names =
-    match Pdf.lookup_direct_orelse pdf "/Names" "/Nums" tree with
-    | Some (Pdf.Array elts) ->
-        if odd (length elts)
-          then
-             begin
-               Pdfe.log "Bad /Names array. Name tree will be read as empty\n";
-               []
-             end
-          else pairs_of_list elts
-    | _ -> []
-  in
-    match Pdf.lookup_direct pdf "/Kids" tree with
-    | Some (Pdf.Array kids) ->
-        names @ flatten (map (read_name_tree pdf) kids)
-    | _ -> names
-
-let read_number_tree pdf tree =
-  let r = read_name_tree pdf tree in
-    try
-      map (function (Pdf.Integer i, x) -> (string_of_int i, x) | _ -> raise Exit) r
-    with
-      Exit ->
-        Pdfe.log "Pdfmerge.read_number_tree: skipping malformed name tree\n";
-        []
-
-let read_name_tree pdf tree =
-  let r = read_name_tree pdf tree in
-    try
-      map (function (Pdf.String s, x) -> (s, x) | _ -> raise Exit) r
-    with
-      Exit ->
-        Pdfe.log "Pdfmerge.read_name_tree: skipping malformed name tree\n";
-        []
-
-let maxsize = 10 (* Must be at least two *)
-
-type ('k, 'v) nt =
-  Br of 'k * ('k, 'v) nt list * 'k
-| Lf of 'k * ('k * 'v) list * 'k
-
-let left l = fst (hd l)
-let right l = fst (last l)
-
-let rec build_nt_tree l =
-  if length l = 0 then assert false;
-  if length l <= maxsize
-    then Lf (left l, l, right l)
-    else Br (left l, map build_nt_tree (splitinto (length l / maxsize) l), right l)
-
-let rec name_tree_of_nt isnum isroot pdf = function
-  Lf (llimit, items, rlimit) ->
-    let entry =
-      [((if isnum then "/Nums" else "/Names"),
-        Pdf.Array (flatten (map (fun (k, v) -> [(if isnum then Pdf.Integer (int_of_string k) else Pdf.String k); v]) items)))] 
-    in
-    let ll, rl =
-      if isnum
-        then Pdf.Integer (int_of_string llimit), Pdf.Integer (int_of_string rlimit)
-        else Pdf.String llimit, Pdf.String rlimit
-    in
-      Pdf.Dictionary
-        (entry @
-         if isroot then [] else [("/Limits", Pdf.Array [ll; rl])])
-| Br (llimit, nts, rlimit) ->
-    let indirects =
-      let kids = map (name_tree_of_nt isnum false pdf) nts in
-        map (Pdf.addobj pdf) kids
-    in
-    let ll, rl =
-      if isnum
-        then Pdf.Integer (int_of_string llimit), Pdf.Integer (int_of_string rlimit)
-        else Pdf.String llimit, Pdf.String rlimit
-    in
-      Pdf.Dictionary
-       [("/Kids", Pdf.Array (map (fun x -> Pdf.Indirect x) indirects));
-        ("/Limits", Pdf.Array [ll; rl])]
-
-let build_name_tree isnum pdf = function
-  | [] ->
-      if isnum then
-        Pdf.Dictionary [("/Nums", Pdf.Array [])]
-      else
-        Pdf.Dictionary [("/Names", Pdf.Array [])]
-  | ls ->
-      let nt = build_nt_tree (sort compare ls) in
-        name_tree_of_nt isnum true pdf nt
-
-(* Once we know there are no clashes *)
-let merge_name_trees_no_clash pdf trees =
-  build_name_tree false pdf (flatten (map (read_name_tree pdf) trees))
-
-let merge_number_trees_no_clash pdf trees =
-  build_name_tree true pdf (flatten (map (read_number_tree pdf) trees))
-
 (* Merging entries in the Name Dictionary. [pdf] here is the new merged pdf, [pdfs] the original ones. *)
 let merge_namedicts pdf pdfs =
   let names =
@@ -206,7 +108,7 @@ let merge_namedicts pdf pdfs =
       in
         let new_trees =
           map
-            (fun (name, trees) -> name, merge_name_trees_no_clash pdf trees)
+            (fun (name, trees) -> name, Pdftree.merge_name_trees_no_clash pdf trees)
             with_names
         in
           (* Add all the trees as indirect references to the pdf *)
@@ -237,14 +139,14 @@ let merge_bookmarks changes pdfs ranges pdf =
       let catalog = Pdf.catalog_of_pdf pdf in
         let oldstyle =
           match Pdf.lookup_direct pdf "/Dests" catalog with
-          | Some d -> read_name_tree pdf d
+          | Some d -> Pdftree.read_name_tree pdf d
           | _ -> []
         in
         let newstyle =
           match Pdf.lookup_direct pdf "/Names" catalog with
           | Some d ->
               begin match Pdf.lookup_direct pdf "/Dests" d with
-              | Some d -> read_name_tree pdf d
+              | Some d -> Pdftree.read_name_tree pdf d
               | _ -> []
               end
           | _ -> []
@@ -490,7 +392,7 @@ let merge_pdfs_rename_name_trees names pdfs =
   let names =
     map
       (function
-       | (pdf, Some nametree) -> (pdf, map fst (read_name_tree pdf nametree))
+       | (pdf, Some nametree) -> (pdf, map fst (Pdftree.read_name_tree pdf nametree))
        | (pdf, None) -> (pdf, []))
       pdfs_and_nametrees
   in
@@ -624,116 +526,17 @@ let merge_acroforms pdf pdfs =
     in
       Some (Pdf.addobj pdf new_dict)
 
-(* Merge structure hierarchy / tagged PDF.
-
-/IDTree                 name tree     merge
-/ParentTree             number tree   merge
-/ParentTreeNextKey      integer       delete for now, since optional and needs updating
-/RoleMap                dict          merge
-/ClassMap               dict          merge
-/NameSpaces             array         merge
-/PronunciationLexicon   array         merge
-/AF                     array         merge
-/K                      dict/array    merge
-*)
-let merge_structure_hierarchy pdf pdfs =
-  let get_all struct_tree_roots pdf name =
-    option_map
-      (fun str -> Pdf.lookup_direct pdf name str) struct_tree_roots
-  in
-  let merge_dicts dicts =
-    fold_left
-      (fun d (k, v) -> Pdf.add_dict_entry d k v)
-      (Pdf.Dictionary [])
-      (flatten
-        (option_map
-          (function
-           | Pdf.Dictionary d -> Some d
-           | _ -> Pdfe.log "merge_dicts: not a dict"; None) dicts))
-  in
-  let merge_arrays arrays =
-    Pdf.Array
-      (flatten
-        (option_map
-          (function
-           | Pdf.Array a -> Some a
-           | _ -> Pdfe.log "merge_array: not an array"; None) arrays))
-  in
-  let mkarray = function
-    | Pdf.Array a -> Pdf.Array a
-    | x -> Pdf.Array [x]
-  in
-  let struct_tree_roots, struct_tree_objnums =
-    split
-      (option_map
-        (fun pdf ->
-          let catalog = Pdf.catalog_of_pdf pdf in
-             match Pdf.lookup_direct pdf "/StructTreeRoot" catalog with
-             | None -> None
-             | Some str ->
-                Some
-                  (str,
-                   match catalog with
-                   | Pdf.Dictionary d ->
-                       begin match lookup "/StructTreeRoot" d with Some (Pdf.Indirect i) -> i | _ -> 0 end
-                   | _ -> raise (Pdf.PDFError "merge_structure_hierarchy: bad catalog")))
-        pdfs)
-  in
-    match struct_tree_roots with
-    | [] -> None
-    | [x] ->
-       Some (hd struct_tree_objnums) (* if only one, don't interfere, just preserve it. *)
-    | _ ->
-      let merged_idtree =
-        merge_name_trees_no_clash pdf (get_all struct_tree_roots pdf "/IDTree")
-      in
-      let merged_parenttree =
-        merge_number_trees_no_clash pdf (get_all struct_tree_roots pdf "/ParentTree")
-      in
-      let merged_rolemap =
-        merge_dicts (get_all struct_tree_roots pdf "/RoleMap") in
-      let merged_classmap =
-        merge_dicts (get_all struct_tree_roots pdf "/ClassMap") in
-      let merged_namespaces =
-        merge_arrays (get_all struct_tree_roots pdf "/NameSpaces") in
-      let merged_pronunciation_lexicon =
-        merge_arrays (get_all struct_tree_roots pdf "/PronunciationLexicon") in
-      let merged_af =
-        merge_arrays (get_all struct_tree_roots pdf "/AF") in
-      let struct_tree_objnum = Pdf.addobj pdf Pdf.Null in
-      let merged_k =
-        match merge_arrays (map mkarray (get_all struct_tree_roots pdf "/K")) with
-        | Pdf.Array l ->
-            Pdf.Array (map (fun d -> Pdf.add_dict_entry d "/P" (Pdf.Indirect struct_tree_objnum)) (map (Pdf.direct pdf) l))
-        | _ -> assert false
-      in
-        let optional n = function
-        | Pdf.Dictionary [] -> []
-        | Pdf.Array [] -> []
-        | x -> [(n, Pdf.Indirect (Pdf.addobj pdf x))]
-        in
-        let new_dict =
-          Pdf.Dictionary
-            (["/Type", Pdf.Name "/StructTreeRoot"]
-             @ optional "/IDTree" merged_idtree
-             @ optional "/ParentTree" merged_parenttree
-             @ optional "/RoleMap" merged_rolemap
-             @ optional "/ClassMap" merged_classmap
-             @ optional "/NameSpaces" merged_namespaces
-             @ optional "/PronunciationLexicon" merged_pronunciation_lexicon
-             @ optional "/AF" merged_af
-             @ optional "/K" merged_k)
-        in
-          Pdf.addobj_given_num pdf (struct_tree_objnum, new_dict);
-          Some struct_tree_objnum
-
 let max_version_number pdfs =
   if pdfs = [] then raise (Invalid_argument "max_version_number") else
     hd (sort compare (map (fun p -> (p.Pdf.major, p.Pdf.minor)) pdfs))
 
-let merge_pdfs retain_numbering do_remove_duplicate_fonts ?(struct_hierarchy=true) names pdfs ranges =
+let merge_pdfs retain_numbering do_remove_duplicate_fonts ?(process_struct_trees=true) names pdfs ranges =
+  if process_struct_trees && length (setify names) < length names then
+    Pdfe.log "Warning: multiply-included files will not merge structure trees properly.\n";
+  if process_struct_trees then iter2 Pdfst.trim_structure_tree pdfs ranges;
   let pdfs = merge_pdfs_renumber names pdfs in
     merge_pdfs_rename_name_trees names pdfs;
+    Pdfst.renumber_parent_trees pdfs;
     let major', minor' = max_version_number pdfs in
       let pagelists = map Pdfpage.pages_of_pagetree pdfs
       in let pdf = Pdf.empty () in
@@ -780,8 +583,8 @@ let merge_pdfs retain_numbering do_remove_duplicate_fonts ?(struct_hierarchy=tru
                   extra_catalog_entries
             in
             let extra_catalog_entries =
-              if not struct_hierarchy then extra_catalog_entries else
-                match merge_structure_hierarchy pdf pdfs with
+              if not process_struct_trees then extra_catalog_entries else
+                match Pdfst.merge_structure_trees pdf pdfs with
                 | None -> extra_catalog_entries
                 | exception e ->
                     Pdfe.log (Printf.sprintf "Warning: failed to merge structure tree roots (%s)\n" (Printexc.to_string e));
