@@ -241,7 +241,8 @@ let endpage_fast pdf =
     | _ -> 0
 
 let _ =
-  Pdfread.endpage := endpage_fast
+  Pdfread.endpage := endpage;
+  Pdfst.endpage := endpage
 
 (* Make a collection of pages capable of being merged -- in other words rename
 their resources so as not to clash. *)
@@ -967,7 +968,7 @@ let fixup_parents pdf =
       Some pagetreeroot -> fixup_parents_inner pdf 0 pagetreeroot
     | _ -> raise (Pdf.PDFError "fixup_parents: no page tree root")
 
-let pdf_of_pages ?(retain_numbering = false) basepdf range =
+let pdf_of_pages ?(retain_numbering = false) ?(process_struct_tree = false) basepdf range =
   let page_labels =
     if length (Pdfpagelabels.read basepdf) = 0 then [] else
       if retain_numbering
@@ -1023,90 +1024,103 @@ let pdf_of_pages ?(retain_numbering = false) basepdf range =
           {pdf with
              Pdf.major = basepdf.Pdf.major;
              Pdf.minor = basepdf.Pdf.minor;
+             Pdf.root = basepdf.Pdf.root; (* So structure tree trimmer can run *)
+             Pdf.was_linearized = basepdf.Pdf.was_linearized;
              Pdf.trailerdict = basepdf.Pdf.trailerdict;
              Pdf.saved_encryption = basepdf.Pdf.saved_encryption}
         in
+          (*Printf.eprintf "----BEFORE trim_structure_tree\n";
+          Pdfwrite.debug_whole_pdf pdf; *)
+          if process_struct_tree then Pdfst.trim_structure_tree pdf range;
+          (*Printf.eprintf "----AFTER trim_structure_tree\n";
+          Pdfwrite.debug_whole_pdf pdf; *)
           let existing_root_entries =
             try
               match Pdf.lookup_obj basepdf basepdf.Pdf.root with | Pdf.Dictionary d -> d | _ -> []
             with
               _ -> []
           in
-              let objnumbers = map (function i -> select i page_numbers) range in
-              let old_pagetree_root_num =
-                match Pdf.lookup_direct basepdf "/Root" pdf.Pdf.trailerdict with
-                | Some (Pdf.Dictionary d) ->
-                    begin match lookup "/Pages" d with
-                    | Some (Pdf.Indirect i) -> i
-                    | _ -> raise (Pdf.PDFError "pdf_of_pages")
+            let objnumbers = map (function i -> select i page_numbers) range in
+            let old_pagetree_root_num =
+              match Pdf.lookup_direct basepdf "/Root" pdf.Pdf.trailerdict with
+              | Some (Pdf.Dictionary d) ->
+                  begin match lookup "/Pages" d with
+                  | Some (Pdf.Indirect i) -> i
+                  | _ -> raise (Pdf.PDFError "pdf_of_pages")
+                  end
+              | _ -> raise (Pdf.PDFError "pdf_of_pages")
+            in
+            (* 1. Look through all the page objects to be included, and
+            replicate inheritable entries from their parent nodes, since they
+            may fail to exist, leaving pages without Media boxes or
+            resources! Inheritable entries are /MediaBox /CropBox /Rotate
+            /Resources *)
+            iter
+              (function objnum ->
+                 let replace_inherit objnum entry =
+                   let obj = Pdf.lookup_obj pdf objnum in
+                     (* Find the first parent entry we can which has the correct attribute. *)
+                     let rec find_attribute obj =
+                       (* Only replace if not there! *)
+                       match Pdf.lookup_direct pdf entry obj with
+                       | Some _ -> None
+                       | _ ->
+                         match Pdf.lookup_direct pdf "/Parent" obj with
+                         | Some (Pdf.Dictionary parent) ->
+                             (* Does this one have the attribute? If yes,
+                             return, if no carry on looking... Don't use a
+                             direct lookup, because we want to retain the
+                             indirect reference to resources, for example. *)
+                             begin match lookup entry parent with
+                             | Some pdfobj -> Some pdfobj
+                             | None -> find_attribute (Pdf.Dictionary parent)
+                             end
+                         | _ -> None (* Got to top, couldn't find anything *)
+                     in
+                       match find_attribute obj with
+                       | None -> ()
+                       | Some replacement_attr ->
+                          (* Replace the attribute with replacement_attr, updating the page object in place. *)
+                          Pdf.addobj_given_num pdf (objnum, Pdf.add_dict_entry obj entry replacement_attr)
+                 in
+                   replace_inherit objnum "/MediaBox";
+                   replace_inherit objnum "/CropBox";
+                   replace_inherit objnum "/Rotate";
+                   replace_inherit objnum "/Resources")
+              objnumbers;
+            let thetree = pagetree_with_objnumbers true old_pagetree_root_num (source pdf.Pdf.objects.Pdf.maxobjnum) objnumbers 0 in
+            (* 2. Kill the old page tree, excepting pages which will appear in the new
+            PDF. It will link, via /Parent entries etc, to the new page tree. To do
+            this, we remove all objects with /Type /Page or /Type /Pages. The other
+            places that null can appear, in destinations and so on, are ok, we think. *)
+            Pdf.objiter
+              (fun i o ->
+                match o with
+                | Pdf.Dictionary d ->
+                    begin match lookup "/Type" d with
+                    | Some (Pdf.Name ("/Pages")) -> Pdf.removeobj pdf i
+                    | Some (Pdf.Name ("/Page")) ->
+                        if not (mem i objnumbers) then Pdf.removeobj pdf i
+                    | _ -> ()
                     end
-                | _ -> raise (Pdf.PDFError "pdf_of_pages")
-              in
-              (* 1. Look through all the page objects to be included, and
-              replicate inheritable entries from their parent nodes, since they
-              may fail to exist, leaving pages without Media boxes or
-              resources! Inheritable entries are /MediaBox /CropBox /Rotate
-              /Resources *)
-              iter
-                (function objnum ->
-                   let replace_inherit objnum entry =
-                     let obj = Pdf.lookup_obj pdf objnum in
-                       (* Find the first parent entry we can which has the correct attribute. *)
-                       let rec find_attribute obj =
-                         (* Only replace if not there! *)
-                         match Pdf.lookup_direct pdf entry obj with
-                         | Some _ -> None
-                         | _ ->
-                           match Pdf.lookup_direct pdf "/Parent" obj with
-                           | Some (Pdf.Dictionary parent) ->
-                               (* Does this one have the attribute? If yes,
-                               return, if no carry on looking... Don't use a
-                               direct lookup, because we want to retain the
-                               indirect reference to resources, for example. *)
-                               begin match lookup entry parent with
-                               | Some pdfobj -> Some pdfobj
-                               | None -> find_attribute (Pdf.Dictionary parent)
-                               end
-                           | _ -> None (* Got to top, couldn't find anything *)
-                       in
-                         match find_attribute obj with
-                         | None -> ()
-                         | Some replacement_attr ->
-                            (* Replace the attribute with replacement_attr, updating the page object in place. *)
-                            Pdf.addobj_given_num pdf (objnum, Pdf.add_dict_entry obj entry replacement_attr)
-                   in
-                     replace_inherit objnum "/MediaBox";
-                     replace_inherit objnum "/CropBox";
-                     replace_inherit objnum "/Rotate";
-                     replace_inherit objnum "/Resources")
-                objnumbers;
-              let thetree = pagetree_with_objnumbers true old_pagetree_root_num (source pdf.Pdf.objects.Pdf.maxobjnum) objnumbers 0 in
-              (* 2. Kill the old page tree, excepting pages which will appear in the new
-              PDF. It will link, via /Parent entries etc, to the new page tree. To do
-              this, we remove all objects with /Type /Page or /Type /Pages. The other
-              places that null can appear, in destinations and so on, are ok, we think. *)
-              Pdf.objiter
-                (fun i o ->
-                  match o with
-                  | Pdf.Dictionary d ->
-                      begin match lookup "/Type" d with
-                      | Some (Pdf.Name ("/Pages")) -> Pdf.removeobj pdf i
-                      | Some (Pdf.Name ("/Page")) ->
-                          if not (mem i objnumbers) then Pdf.removeobj pdf i
-                      | _ -> ()
-                      end
-                  | _ -> ())
-                pdf;
-                (* Now, add the new page tree, with root at the same object
-                number, and finish *)
-                let new_pagetree = pdf_of_pages_build_pagetree thetree objnumbers pdf in
-                  Pdf.addobj_given_num pdf (old_pagetree_root_num, new_pagetree);
-                    let pdf = add_root old_pagetree_root_num existing_root_entries pdf in
-                    Pdfpagelabels.write pdf page_labels;
-                    let pdf = Pdfmarks.add_bookmarks marks pdf in
-                      fixup_duplicate_pages pdf;
-                      fixup_parents pdf;
-                      pdf
+                | _ -> ())
+              pdf;
+              (* Now, add the new page tree, with root at the same object
+              number, and finish *)
+              let new_pagetree = pdf_of_pages_build_pagetree thetree objnumbers pdf in
+                Pdf.addobj_given_num pdf (old_pagetree_root_num, new_pagetree);
+                  let pdf = add_root old_pagetree_root_num existing_root_entries pdf in
+                  Pdfpagelabels.write pdf page_labels;
+                  let pdf = Pdfmarks.add_bookmarks marks pdf in
+                    fixup_duplicate_pages pdf;
+                    fixup_parents pdf;
+                    (*Printf.eprintf "----BEFORE postprocess_parent_tree\n";
+                    Pdfwrite.debug_whole_pdf pdf; *)
+                    (* Disabled - verifier does not care for this presently... *)
+                    (*if process_struct_tree then Pdfst.postprocess_parent_tree pdf;*)
+                    (*Printf.eprintf "----AFTER postprocess_parent_tree\n";
+                    Pdfwrite.debug_whole_pdf pdf; *)
+                    pdf
 
 let prepend_operators pdf ops ?(fast=false) page =
   if fast then
