@@ -194,7 +194,7 @@ let getuntil_white_or_delimiter =
 let getuntil_white_or_delimiter_string =
   getuntil_string true is_whitespace_or_delimiter
 
-(* \section {Lexing} *)
+(* Lexing *)
 
 (* Each of the following functions lexes a particular object, leaving the
 channel position at the character after the end of the lexeme. Upon entry, the
@@ -216,7 +216,6 @@ let lex_number i =
       | Pdfgenlex.LexReal f -> LexReal f
       | _ -> LexNone
     with
-    (* FIXME: Heed warning 52 *)
     | Failure x when x = "hd" -> LexNone
     | Pdf.PDFError _ (* can't cope with floats with leading point. *)
     | Failure _ (*"int_of_string"*) ->
@@ -499,7 +498,7 @@ let lex_stream_data i l opt =
     with
       _ -> raise (Pdf.PDFError (Pdf.input_pdferror i "lex_stream_data"))
 
-(* Lex a stream. This involves \emph{parsing} the stream dictionary to get the
+(* Lex a stream. This involves parsing the stream dictionary to get the
 length. [i] is at the start of the stream data, suitable for input to
 [lex_stream_data]. We extract the dictionary by going through
 [previous_lexemes], the reverse-order list of the lexemes already read. *)
@@ -595,15 +594,21 @@ let lex_next dictlevel arraylevel endonstream i previous_lexemes p opt lexobj =
 
 (* Lex just a dictionary, consuming only the tokens to the end of it. This is
 used in the [PDFPages] module to read dictionaries in graphics streams. *)
-let lex_dictionary i =
+let lex_dictionary minus_one i =
   let rec lex_dictionary_getlexemes i lexemes dictlevel arraylevel =
     let lex_dictionary_next i dictlevel arraylevel =
       let dummyparse = fun _ -> 0, Pdf.Null
       in let dummylexobj = fun _ -> [] in
+        (*Printf.eprintf "lex_dictionary_next: dictlevel = %i\n" !dictlevel;*)
+        let r =
         lex_next dictlevel arraylevel false i [] dummyparse false dummylexobj
+        in
+          (*Printf.eprintf "Read lexeme %S\n" (string_of_lexeme r);
+          Printf.eprintf "after lex_dictionary_next: dictlevel = %i\n" !dictlevel;*)
+          r
     in
       match lex_dictionary_next i dictlevel arraylevel with
-      | LexRightDict when !dictlevel = 0 && !arraylevel = 0 ->
+      | LexRightDict when !dictlevel = (if minus_one then -1 else 0) && !arraylevel = 0 ->
           rev (LexRightDict::lexemes)
       | StopLexing ->
           rev lexemes
@@ -701,16 +706,15 @@ let lex_object_at oneonly i opt p lexobj =
     in
       lex_object_at i []
 
-
 (* Type of sanitized cross-reference entries. They are either plain offsets, or
 an object stream an index into it. *)
 type xref =
   | XRefPlain of int * int (* offset, generation. *)
   | XRefStream of int * int (* object number of stream, index. *)
 
-let string_of_xref = function
+(*let string_of_xref = function
   | XRefPlain (p, i) -> Printf.sprintf "XRefPlain (%i, %i)" p i
-  | XRefStream (o, i) -> Printf.sprintf "XrefStream %i, index %i" o i
+  | XRefStream (o, i) -> Printf.sprintf "XrefStream %i, index %i" o i*)
 
 let xrefs_table_create () = Hashtbl.create 1001
 
@@ -754,7 +758,7 @@ type partial_parse_element =
   | Lexeme of Pdfgenlex.t
   | Parsed of Pdf.pdfobject
 
-let print_parseme = function
+(*let print_parseme = function
   | Parsed p ->
       flprint "PARSED:";
       print_string (Pdfwrite.string_of_pdf p);
@@ -762,7 +766,7 @@ let print_parseme = function
   | Lexeme l ->
       flprint "LEXEME:";
       print_lexeme l;
-      flprint "\n" 
+      flprint "\n"*)
 
 (* Parse stage one. *)
 let parse_R ts =
@@ -836,7 +840,7 @@ and parse_array sofar = function
       parse_array (h::sofar) t
 
 (* Main function *)
-let rec parse_to_tree (sofar : partial_parse_element list) = function
+let rec parse_to_tree sofar = function
   | Lexeme LexLeftDict::t ->
       let dict, rest = parse_dictionary [] t in
         parse_to_tree (dict::sofar) rest
@@ -910,13 +914,7 @@ let parse ?(failure_is_ok = false) lexemes =
       if failure_is_ok then (max_int, Pdf.Null) else raise e
 
 let parse_single_object s =
-  snd (parse
-    (lex_object_at
-       true
-       (Pdfio.input_of_string s)
-       true
-       (fun _ -> (0, Pdf.Null))
-       (fun _ -> [])))
+  snd (parse (lex_object_at true (Pdfio.input_of_string s) true (fun _ -> (0, Pdf.Null)) (fun _ -> [])))
 
 (* Given an object stream pdfobject and a list of object indexes to extract,
  return an [(int * Pdf.objectdata) list] representing those object number,
@@ -924,12 +922,12 @@ let parse_single_object s =
 let lex_stream_object
   i xrefs parse opt obj indexes user_pw owner_pw partial_pdf gen
 =
-  (*if !read_debug then
+  if !read_debug then
     begin
       Pdfe.log (Printf.sprintf "lexing object stream at %i\n" (i.Pdfio.pos_in ()));
       Pdfe.log (Printf.sprintf "lexing object stream %i\nTo find the indexes:\n" obj);
-      iter (Pdfe.log (Printf.sprintf "%i ")) indexes; Pdfe.log "\n"
-    end;*)
+      iter (fun i -> Pdfe.log (Printf.sprintf "%i " i)) indexes; Pdfe.log "\n"
+    end;
   let _, stmobj = parse (lex_object i xrefs parse opt obj) in
     match stmobj with
     | Pdf.Stream {contents = d, stream} ->
@@ -994,8 +992,6 @@ let lex_stream_object
                            end;
                            incr index)
                         pairs;
-                      (* FIXME: Why does this make it null rather than actually
-                       * remove the object? *)
                       Pdf.addobj_given_num partial_pdf (obj, Pdf.Null);
                       !objects
                 with
@@ -1015,15 +1011,6 @@ let lex_stream_object
             (Pdf.input_pdferror i (Printf.sprintf "lex_stream_object: not a stream, but %s"
                                    (Pdfwrite.string_of_pdf stmobj))))
 
-(* Advance to the first thing after the current pointer which is not a comment.
-*)
-let rec ignore_comments i =
-  let pos = i.pos_in () in
-    match i.input_char () with
-    | Some '%' -> ignore (input_line i); ignore_comments i
-    | Some _ -> i.seek_in pos
-    | None -> raise End_of_file
-
 (* Cross-reference tables *)
 
 (* Read the cross-reference table. Supports the multiple sections created when
@@ -1296,7 +1283,7 @@ let get_object i xrefs n =
 let is_linearized i =
   try
     ignore (read_header i);
-    let lexemes = lex_dictionary i in
+    let lexemes = lex_dictionary false i in
       let _, parsed = parse ~failure_is_ok:true lexemes in
         match Pdf.lookup_direct (Pdf.empty ()) "/Linearized" parsed with
         | Some (Pdf.Integer _ | Pdf.Real _) -> true
@@ -1652,9 +1639,19 @@ let read_pdf ?revision user_pw owner_pw opt i =
             (* Delete items in !postdeletes - these are any xref streams, and
             object streams (if finished with).  This allows decryption to be
             performed without accidently trying to decrypt these streams -
-            which would fail due to them not being of set lengths. Also saves
-            memory *)
-            iter (Pdf.removeobj pdf) !postdeletes;
+            which would fail due to them not being of set lengths. *)
+            iter
+              (fun i ->
+                 (* Oct 2023 - check it's not been overwritten by a later revision of the file *)
+                 match Pdf.lookup_obj pdf i with
+                 | Pdf.Stream {contents = (Pdf.Dictionary d, _)}  ->
+                    begin match lookup "/Type" d with
+                    | Some (Pdf.Name ("/XRef" | "/ObjStm")) -> Pdf.removeobj pdf i
+                    | _ -> ()
+                    end
+                 | exception Not_found -> ()
+                 | _ -> ())
+              !postdeletes;
             (* Add stream_object_ids *)
             pdf.Pdf.objects.Pdf.object_stream_ids <- object_stream_ids;
             (*if !read_debug then
@@ -1776,7 +1773,16 @@ let read_malformed_pdf_objects i =
         with
           e ->
             if !read_debug then Pdfe.log "Couldn't get object, moving on\n";
-            ignore (input_line i) (* Move on *)
+            (* 26/02/2024 advance only to any line beginning or ending with
+               'endobj'. Before, a runaway string read, for example, could
+               advance too far, perhaps into the middle of a stream, then no
+               more objects could be read. *)
+            i.seek_in c;
+            let fin = ref false in
+              while not !fin do
+                let l = String.trim (input_line i) in
+                  fin := starts_with "endobj" l || starts_with "jbodne" (implode (rev (explode l)))
+              done
     done;
     !objs
 
